'use strict';

exports.__esModule = true;

var _referenceContext = require('reference-context');

var _referenceContext2 = _interopRequireDefault(_referenceContext);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _possibleConstructorReturn(self, call) { if (!self) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return call && (typeof call === "object" || typeof call === "function") ? call : self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function, not " + typeof superClass); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, enumerable: false, writable: true, configurable: true } }); if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass; } /**
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * Copyright 2017 Moshe Simantov
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                *
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * Licensed under the Apache License, Version 2.0 (the "License");
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * you may not use this file except in compliance with the License.
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * You may obtain a copy of the License at
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                *
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                *     http://www.apache.org/licenses/LICENSE-2.0
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                *
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * Unless required by applicable law or agreed to in writing, software
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * distributed under the License is distributed on an "AS IS" BASIS,
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * See the License for the specific language governing permissions and
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * limitations under the License.
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                */

/**
 * @alias ReferenceContext
 */


// Maximum number of references allowed on this context, which is about 65K references.
// That's already more the enough, but can be changed up to `Number.MAX_SAFE_INTEGER`.
var MAX_REFERENCES_DEFAULT = Math.pow(256, 2);

var kRefNumMax = Symbol('refNumMax');
var kNextRefId = Symbol('nextRefId');

/**
 * A {@link ReferenceContext} with assignment support.
 *
 * @extends {ReferenceContext}
 */

var AssignableContext = function (_ReferenceContext) {
  _inherits(AssignableContext, _ReferenceContext);

  /**
   * @param {object} context The context
   * @param {ReferenceContext|null} parentContext The context parent
   * @param {object} opts Options
   * @param {number} [opts.maxReferences=65536] The maximum references allowed to assign on this
   * context
   */
  function AssignableContext(context, parentContext) {
    var opts = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

    _classCallCheck(this, AssignableContext);

    var _this = _possibleConstructorReturn(this, _ReferenceContext.call(this, {}, parentContext));

    if (parentContext instanceof AssignableContext) {
      /**
       * @type {number}
       * @private
       */
      _this[kRefNumMax] = parentContext[kRefNumMax];

      /**
       * @type {number}
       * @private
       */
      _this[kNextRefId] = parentContext[kNextRefId];
    } else {
      _this[kRefNumMax] = opts.maxReferences || MAX_REFERENCES_DEFAULT;
      _this[kNextRefId] = 1;
    }

    _this.copyFrom(context);
    return _this;
  }

  /**
   * @override
   */


  AssignableContext.prototype.generateReference = function generateReference() {
    // `this.size` is not the real generated reference number,
    // but if we got so many references it's better to stop now.
    if (this.size >= this[kRefNumMax]) {
      throw new Error('Maximum references number reached: ' + this[kRefNumMax]);
    }

    var refId = this[kNextRefId];

    if (this[kNextRefId] < Number.MAX_SAFE_INTEGER) {
      this[kNextRefId] += 1;
    } else {
      this[kNextRefId] = 1;
    }

    return refId;
  };

  return AssignableContext;
}(_referenceContext2.default);

exports.default = AssignableContext;