'use strict';

exports.__esModule = true;

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

var _AssignableContext2 = require('./AssignableContext');

var _AssignableContext3 = _interopRequireDefault(_AssignableContext2);

var _EnvContext = require('./EnvContext');

var _EnvContext2 = _interopRequireDefault(_EnvContext);

var _RemoteSession = require('./RemoteSession');

var _RemoteSession2 = _interopRequireDefault(_RemoteSession);

var _parser = require('./parser');

var _parser2 = _interopRequireDefault(_parser);

var _LocalPromise = require('./LocalPromise');

var _LocalPromise2 = _interopRequireDefault(_LocalPromise);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _possibleConstructorReturn(self, call) { if (!self) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return call && (typeof call === "object" || typeof call === "function") ? call : self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function, not " + typeof superClass); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, enumerable: false, writable: true, configurable: true } }); if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass; } /**
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * Copyright 2017 Moshe Simantov
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                *
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * Licensed under the Apache License, Version 2.0 (the "License");
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * you may not use this file except in compliance with the License.
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * You may obtain a copy of the License at
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                *
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                *     http://www.apache.org/licenses/LICENSE-2.0
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                *
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * Unless required by applicable law or agreed to in writing, software
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * distributed under the License is distributed on an "AS IS" BASIS,
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * See the License for the specific language governing permissions and
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * limitations under the License.
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                */

if (!('getOwnPropertyDescriptors' in Object)) {
  throw new Error('Old versions should require "babel-polyfill" module.');
}

var kChildRefs = Symbol('childRefs');

function getPropValueSafe(obj, prop) {
  try {
    return obj[prop];
  } catch (err) {
    return undefined;
  }
}

/**
 * @example
 * import { Context } from 'remote-context';
 *
 * const context = new Context(envContext);
 */

var Context = function (_AssignableContext) {
  _inherits(Context, _AssignableContext);

  /**
   * Check if the given value is unique per this environment.
   * For example the string `'foo'` is not referable, but the object `{ foo: 1 }` is referable.
   *
   * @param {*} value The given value to check
   * @return {boolean} True if the given value is referable
   */
  Context.isReferable = function isReferable(value) {
    switch (typeof value === 'undefined' ? 'undefined' : _typeof(value)) {
      case 'function':
      case 'symbol':
        return true;

      case 'object':
        return value !== null;

      default:
        return false;
    }
  };

  /**
   * Create a new virtual context on the given environment.
   *
   * @param {EnvContext} envContext The given environment
   * @param {object} context The context to create
   * @param {object} opts Options for {@link AssignableContext.constructor}
   */


  function Context(envContext) {
    var context = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
    var opts = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

    _classCallCheck(this, Context);

    if (!(envContext instanceof _EnvContext2.default)) {
      throw new TypeError('Invalid "envContext" param: ' + envContext);
    }

    /**
     * @type {Map}
     * @private
     */
    var _this = _possibleConstructorReturn(this, _AssignableContext.call(this, {}, envContext, opts));

    _this[kChildRefs] = new Map();

    _this.copyFrom(context);
    return _this;
  }

  /**
   * Get the remote context of this context.
   * A single context can be served to multiple peers, each peer will get a separate copy of the
   * context.
   *
   * @param {Stream} stream a duplex stream to the remote peer context
   * @param {object} [opts] Options to {@link RemoteSession}
   * @return {RemoteContext}
   */


  Context.prototype.remote = function remote(stream) {
    var opts = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

    var objectStream = opts.objectMode !== true ? _parser2.default.transform(stream) : stream;

    var session = new _RemoteSession2.default(objectStream, this, opts);
    return session.remote;
  };

  /**
   * Fetch a local reference and return a local promise.
   *
   * @param {Reference} reference The local reference to fetch
   * @return {LocalPromise}
   */


  Context.prototype.fetch = function fetch(reference) {
    try {
      return new _LocalPromise2.default(this.get(reference));
    } catch (error) {
      return new _LocalPromise2.default(error, true);
    }
  };

  /**
   * Resolve a local value and return a local promise to that value.
   *
   * @param {*} value The local value to resolve
   * @return {LocalPromise}
   */
  // eslint-disable-next-line class-methods-use-this


  Context.prototype.resolve = function resolve(value) {
    return new _LocalPromise2.default(value);
  };

  /**
   * @override
   */


  Context.prototype.set = function set(reference, value) {
    var _this2 = this;

    if (typeof reference !== 'string' && typeof reference !== 'number') {
      throw new TypeError('Only string or number references are allowed: ' + (typeof reference === 'undefined' ? 'undefined' : _typeof(reference)));
    }

    _AssignableContext.prototype.set.call(this, reference, value);

    var childRefs = this[kChildRefs].get(value);
    if (!childRefs) {
      childRefs = new Set();
      this[kChildRefs].set(value, childRefs);
    }

    var proto = Object.getPrototypeOf(value);
    if (proto !== null && !this.exists(proto)) {
      var childRef = this.assign(proto);
      childRefs.add(childRef);
    }

    Object.getOwnPropertyNames(value).forEach(function (propName) {
      var propValue = getPropValueSafe(value, propName);
      if (!_this2.constructor.isReferable(propValue) || _this2.exists(propValue)) return;

      var childRef = _this2.assign(propValue);
      childRefs.add(childRef);
    });

    Object.getOwnPropertySymbols(value).forEach(function (symbol) {
      var propValue = getPropValueSafe(value, symbol);
      if (!_this2.constructor.isReferable(propValue) || _this2.exists(propValue) || !_this2.exists(symbol)) return;

      var symbolReference = _this2.lookup(symbol);
      if (!symbolReference) return;

      var childRef = _this2.assign(propValue);
      childRefs.add(childRef);
    });

    return this;
  };

  /**
   * @override
   */


  Context.prototype.release = function release(value) {
    var _this3 = this;

    var ret = _AssignableContext.prototype.release.call(this, value);

    var childRefs = this[kChildRefs].get(value);
    if (childRefs) {
      childRefs.forEach(function (ref) {
        return _this3.delete(ref);
      });
      this[kChildRefs].delete(value);
    }

    return ret;
  };

  /**
   * @override
   */


  Context.prototype.clear = function clear() {
    _AssignableContext.prototype.clear.call(this);
    this[kChildRefs].clear();
  };

  return Context;
}(_AssignableContext3.default);

exports.default = Context;