'use strict';

exports.__esModule = true;

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _AssignableContext2 = require('./AssignableContext');

var _AssignableContext3 = _interopRequireDefault(_AssignableContext2);

var _RemoteSession = require('./RemoteSession');

var _RemoteSession2 = _interopRequireDefault(_RemoteSession);

var _actions = require('./actions');

var _LocalReference = require('./LocalReference');

var _LocalReference2 = _interopRequireDefault(_LocalReference);

var _ObjectSnapshot = require('./ObjectSnapshot');

var _ObjectSnapshot2 = _interopRequireDefault(_ObjectSnapshot);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _possibleConstructorReturn(self, call) { if (!self) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return call && (typeof call === "object" || typeof call === "function") ? call : self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function, not " + typeof superClass); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, enumerable: false, writable: true, configurable: true } }); if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass; } /**
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * Copyright 2017 Moshe Simantov
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                *
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * Licensed under the Apache License, Version 2.0 (the "License");
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * you may not use this file except in compliance with the License.
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * You may obtain a copy of the License at
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                *
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                *     http://www.apache.org/licenses/LICENSE-2.0
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                *
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * Unless required by applicable law or agreed to in writing, software
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * distributed under the License is distributed on an "AS IS" BASIS,
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * See the License for the specific language governing permissions and
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * limitations under the License.
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                */

var kSession = Symbol('session');
var kSessionListener = Symbol('sessionListener');
var kDispatchSet = Symbol('dispatchSet');
var kDispatchSnapshots = Symbol('dispatchSnapshots');

/**
 * A local context that will represent the remote context of the other peer.
 */

var LocalContext = function (_AssignableContext) {
  _inherits(LocalContext, _AssignableContext);

  /**
   * @param {RemoteSession} session
   * @param {ReferenceContext} parentContext
   * @param {object} opts Options for {@link AssignableContext.constructor}
   */
  function LocalContext(session, parentContext) {
    var opts = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

    _classCallCheck(this, LocalContext);

    if (!(session instanceof _RemoteSession2.default)) {
      throw new TypeError('Expect session to be instance of RemoteSession: ' + session);
    }

    /**
     * @type {RemoteSession}
     * @private
     */
    var _this = _possibleConstructorReturn(this, _AssignableContext.call(this, {}, parentContext, opts));

    _this[kSession] = session;

    /**
     * @type {function}
     * @private
     * @return {void}
     */
    _this[kSessionListener] = function () {
      session.removeListener('close', _this[kSessionListener]);

      _this[kSessionListener] = null;
      _this[kSession] = null;

      _this.destroy();
    };
    session.on('close', _this[kSessionListener]);

    /**
     * @type {Set}
     * @private
     */
    _this[kDispatchSet] = new Set();

    /**
     * @type {Map}
     * @private
     */
    _this[kDispatchSnapshots] = new Map();
    return _this;
  }

  /**
   * Get the context remote session
   * @type {RemoteSession}
   */


  /**
   * @override
   */
  LocalContext.prototype.closure = function closure() {
    if (!this[kSession]) {
      throw new TypeError('RemoteSession already closed');
    }

    return new this.constructor(this[kSession], this);
  };

  /**
   * Generate a remote action for dispatching the given value remotely.
   *
   * @param {*} value The given value
   * @return {Action}
   */


  LocalContext.prototype.dispatch = function dispatch(value) {
    if (!this[kSession]) {
      throw new TypeError('RemoteSession already closed');
    }

    var reference = this.assign(value);

    if (!this[kDispatchSet].has(reference)) {
      this[kDispatchSet].add(reference);
      this[kDispatchSnapshots].set(value, new _ObjectSnapshot2.default(value));
      return _actions.RemoteSetAction.fromValue(this[kSession], reference, value);
    }

    var localReference = new _LocalReference2.default(this[kSession], reference, value);
    this[kDispatchSnapshots].get(value).update(value, localReference);

    return new _actions.LocalReferenceAction(reference);
  };

  /**
   * @override
   */


  LocalContext.prototype.set = function set(reference, value) {
    this[kDispatchSet].delete(reference);

    return _AssignableContext.prototype.set.call(this, reference, value);
  };

  /**
   * @override
   */


  LocalContext.prototype.delete = function _delete(reference) {
    this[kDispatchSet].delete(reference);
    return _AssignableContext.prototype.delete.call(this, reference);
  };

  /**
   * @override
   */


  LocalContext.prototype.release = function release(value) {
    this[kDispatchSnapshots].delete(value);
    return _AssignableContext.prototype.release.call(this, value);
  };

  /**
   * @override
   */


  LocalContext.prototype.clear = function clear() {
    this[kDispatchSet].clear();
    this[kDispatchSnapshots].clear();

    return _AssignableContext.prototype.clear.call(this);
  };

  /**
   * End the remote session.
   *
   * @return {LocalContext}
   */


  LocalContext.prototype.end = function end() {
    if (this[kSession]) {
      this[kSession].end();
    }

    return this;
  };

  /**
   * Clear the context and destroy the context session only if this context IS NOT created via
   * closure.
   *
   * @return {void}
   */


  LocalContext.prototype.destroy = function destroy() {
    if (this[kSession]) {
      if (this[kSessionListener]) {
        this[kSession].removeListener('close', this[kSessionListener]);
      }

      if (!(this.parent instanceof LocalContext)) {
        this[kSession].destroy();
      }
    }

    delete this[kSession];
    delete this[kSessionListener];

    this.clear();
  };

  _createClass(LocalContext, [{
    key: 'session',
    get: function get() {
      return this[kSession];
    }
  }]);

  return LocalContext;
}(_AssignableContext3.default);

exports.default = LocalContext;