'use strict';

exports.__esModule = true;

var _RemoteSession = require('./RemoteSession');

var _RemoteSession2 = _interopRequireDefault(_RemoteSession);

var _LocalContext = require('./LocalContext');

var _LocalContext2 = _interopRequireDefault(_LocalContext);

var _actions = require('./actions');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } } /**
                                                                                                                                                           * Copyright 2017 Moshe Simantov
                                                                                                                                                           *
                                                                                                                                                           * Licensed under the Apache License, Version 2.0 (the "License");
                                                                                                                                                           * you may not use this file except in compliance with the License.
                                                                                                                                                           * You may obtain a copy of the License at
                                                                                                                                                           *
                                                                                                                                                           *     http://www.apache.org/licenses/LICENSE-2.0
                                                                                                                                                           *
                                                                                                                                                           * Unless required by applicable law or agreed to in writing, software
                                                                                                                                                           * distributed under the License is distributed on an "AS IS" BASIS,
                                                                                                                                                           * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
                                                                                                                                                           * See the License for the specific language governing permissions and
                                                                                                                                                           * limitations under the License.
                                                                                                                                                           */

var kSession = Symbol('session');
var kReference = Symbol('reference');

/**
 * A reference to a local value. Any change to the LocalReference will be send to the remote peer.
 */

var LocalReference = function () {
  /**
   * Create a reference to a local value.
   *
   * @param {RemoteSession} session The reference remote session
   * @param {Reference} reference The local reference id
   */
  function LocalReference(session, reference) {
    _classCallCheck(this, LocalReference);

    if (!(session instanceof _RemoteSession2.default)) {
      throw new TypeError('Expect session to be instance of RemoteSession: ' + session);
    }
    if (!_LocalContext2.default.isValidReference(reference)) {
      throw new TypeError('The given reference is not valid: ' + reference);
    }

    /**
     * @type {RemoteSession}
     * @private
     */
    this[kSession] = session;

    /**
     * @type {Reference}
     * @private
     */
    this[kReference] = reference;
  }

  /**
   * Set the prototype of the reference value.
   *
   * @param {object|function} prototype The new prototype
   * @return {boolean} True on success
   */


  LocalReference.prototype.setPrototypeOf = function setPrototypeOf(prototype) {
    this[kSession].send(_actions.RemoteSetPrototypeOfAction.fromLocal(this[kSession], this[kReference], prototype));

    return true;
  };

  /**
   * prevent extensions on the reference value.
   *
   * @return {boolean} True on success
   */


  LocalReference.prototype.preventExtensions = function preventExtensions() {
    this[kSession].send(_actions.RemotePreventExtensionsAction.fromLocal(this[kSession], this[kReference]));

    return true;
  };

  /**
   * Define a property on the reference value.
   *
   * @param {Property} property The property
   * @param {Descriptor} descriptor The property descriptor
   * @return {boolean} True on success
   */


  LocalReference.prototype.defineProperty = function defineProperty(property, descriptor) {
    this[kSession].send(_actions.RemoteDefinePropertyAction.fromLocal(this[kSession], this[kReference], property, descriptor));

    return true;
  };

  /**
   * Define a property from the reference value.
   *
   * @param {Property} property The property
   * @return {boolean} True on success
   */


  LocalReference.prototype.deleteProperty = function deleteProperty(property) {
    this[kSession].send(_actions.RemoteDeletePropertyAction.fromLocal(this[kSession], this[kReference], property));

    return true;
  };

  return LocalReference;
}();

exports.default = LocalReference;