'use strict';

exports.__esModule = true;

var _deepEqual = require('deep-equal');

var _deepEqual2 = _interopRequireDefault(_deepEqual);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } } /**
                                                                                                                                                           * Copyright 2017 Moshe Simantov
                                                                                                                                                           *
                                                                                                                                                           * Licensed under the Apache License, Version 2.0 (the "License");
                                                                                                                                                           * you may not use this file except in compliance with the License.
                                                                                                                                                           * You may obtain a copy of the License at
                                                                                                                                                           *
                                                                                                                                                           *     http://www.apache.org/licenses/LICENSE-2.0
                                                                                                                                                           *
                                                                                                                                                           * Unless required by applicable law or agreed to in writing, software
                                                                                                                                                           * distributed under the License is distributed on an "AS IS" BASIS,
                                                                                                                                                           * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
                                                                                                                                                           * See the License for the specific language governing permissions and
                                                                                                                                                           * limitations under the License.
                                                                                                                                                           */

var kPrototype = Symbol('prototype');
var kDescriptors = Symbol('descriptors');

var ObjectSnapshot = function () {
  /**
   * Create a snapshot from the given object.
   *
   * @param {object|function} object The given object
   */
  function ObjectSnapshot(object) {
    _classCallCheck(this, ObjectSnapshot);

    if (!(object instanceof Object)) return;

    /**
     * @type {Object}
     * @private
     */
    this[kPrototype] = Object.getPrototypeOf(object);

    /**
     * @type {object}
     * @private
     */
    this[kDescriptors] = Object.getOwnPropertyDescriptors(object);
  }

  /**
   * Update the local reference of this value with a new version of this snapshot value.
   *
   * @param {object|function} target The new version of this value
   * @param {LocalReference} localReference The local reference of this snapshot value
   * @return {void}
   */


  ObjectSnapshot.prototype.update = function update(target, localReference) {
    if (this[kDescriptors] === undefined) return;

    var prototype = Object.getPrototypeOf(target);
    var descriptors = Object.getOwnPropertyDescriptors(target);

    // Update snapshot immediately before any other actions
    var oldPrototype = this[kPrototype];
    this[kPrototype] = prototype;
    var oldDescriptors = this[kDescriptors];
    this[kDescriptors] = descriptors;

    // Update prototype
    if (oldPrototype !== prototype) {
      localReference.setPrototypeOf(prototype);
    }

    // Remove deleted properties
    Object.keys(oldDescriptors).forEach(function (property) {
      var newDesc = descriptors[property];
      if (newDesc === undefined) {
        localReference.deleteProperty(property);
      }
    });

    // Add or update existed properties
    Object.keys(descriptors).forEach(function (property) {
      var newDesc = descriptors[property];
      var oldDesc = oldDescriptors[property];

      if (!(0, _deepEqual2.default)(newDesc, oldDesc, { strict: true })) {
        localReference.defineProperty(property, newDesc);
      }
    });
  };

  return ObjectSnapshot;
}();

exports.default = ObjectSnapshot;