'use strict';

exports.__esModule = true;

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _referenceContext = require('reference-context');

var _referenceContext2 = _interopRequireDefault(_referenceContext);

var _RemoteSession = require('./RemoteSession');

var _RemoteSession2 = _interopRequireDefault(_RemoteSession);

var _RemoteValue = require('./RemoteValue');

var _RemoteValue2 = _interopRequireDefault(_RemoteValue);

var _RemotePromise = require('./RemotePromise');

var _RemotePromise2 = _interopRequireDefault(_RemotePromise);

var _LocalPromise = require('./LocalPromise');

var _LocalPromise2 = _interopRequireDefault(_LocalPromise);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _possibleConstructorReturn(self, call) { if (!self) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return call && (typeof call === "object" || typeof call === "function") ? call : self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function, not " + typeof superClass); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, enumerable: false, writable: true, configurable: true } }); if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass; } /**
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * Copyright 2017 Moshe Simantov
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                *
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * Licensed under the Apache License, Version 2.0 (the "License");
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * you may not use this file except in compliance with the License.
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * You may obtain a copy of the License at
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                *
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                *     http://www.apache.org/licenses/LICENSE-2.0
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                *
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * Unless required by applicable law or agreed to in writing, software
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * distributed under the License is distributed on an "AS IS" BASIS,
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * See the License for the specific language governing permissions and
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * limitations under the License.
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                */

var _require = require('./actions'),
    GetAction = _require.GetAction,
    DeleteAction = _require.DeleteAction,
    RemoteReferenceAction = _require.RemoteReferenceAction;

var kSession = Symbol('session');
var kSessionListener = Symbol('sessionListener');
var kFetchRequests = Symbol('fetchRequests');

/**
 * A remote context that will represent the local context of the other peer.
 * @extends {ReferenceContext}
 * @example
 * import { RemoteContext } from 'remote-context';
 *
 * const remoteContext = new RemoteContext(session, parentContext);
 */

var RemoteContext = function (_ReferenceContext) {
  _inherits(RemoteContext, _ReferenceContext);

  /**
   * Create a remote context with given remote session and environment context.
   *
   * @param {RemoteSession} session The remote session
   * @param {ReferenceContext} parentContext The environment context
   */
  function RemoteContext(session, parentContext) {
    _classCallCheck(this, RemoteContext);

    if (!(session instanceof _RemoteSession2.default)) {
      throw new TypeError('Expect session to be instance of RemoteSession: ' + session);
    }

    /**
     * @type {RemoteSession}
     * @private
     */
    var _this = _possibleConstructorReturn(this, _ReferenceContext.call(this, {}, parentContext));

    _this[kSession] = session;

    /**
     * @type {function}
     * @private
     * @return {void}
     */
    _this[kSessionListener] = function () {
      session.removeListener('close', _this[kSessionListener]);

      _this[kSessionListener] = null;
      _this[kSession] = null;

      _this.destroy();
    };
    session.on('close', _this[kSessionListener]);

    /**
     * @type {Map}
     * @private
     */
    _this[kFetchRequests] = new Map();
    return _this;
  }

  /**
   * Get the context remote session.
   *
   * @type {RemoteSession}
   */


  /**
   * @override
   */
  RemoteContext.prototype.closure = function closure() {
    if (!this[kSession]) {
      throw new TypeError('RemoteSession already closed');
    }

    return new this.constructor(this[kSession], this);
  };

  /**
   * Fetch a remote reference and return a remote promise.
   * If the value already fetched, return a remote promise of the same value.
   *
   * @param {Reference} reference The remote reference to fetch
   * @param {object} [opts] {@link RemotePromise.constructor} options
   * @return {RemotePromise}
   */


  RemoteContext.prototype.fetch = function fetch(reference) {
    var opts = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

    var promise = this[kFetchRequests].get(reference);
    if (promise) return promise;

    promise = new _RemotePromise2.default(this[kSession], new GetAction(reference), opts);
    this[kFetchRequests].set(reference, promise);

    return promise;
  };

  /**
   * Resolve a remote value and return a remote promise to that value.
   * If the value is not a remote value, return a local promise to the same value.
   *
   * @param {*} value The remote value to resolve
   * @param {object} [opts] {@link RemoteContext.fetch} options
   * @return {RemotePromise|LocalPromise}
   */


  RemoteContext.prototype.resolve = function resolve(value) {
    var opts = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

    if (!(value instanceof Object) || !this.exists(value)) {
      return new _LocalPromise2.default(value);
    }

    var reference = this.lookup(value);
    this[kFetchRequests].delete(reference);

    return this.fetch(reference, opts);
  };

  /**
   * Generate a remote action for dispatching the given value remotely.
   *
   * @param {*} value The given value
   * @throws {ReferenceError} If the value is not {@link RemoteContext#exists}
   * @return {Action}
   */


  RemoteContext.prototype.dispatch = function dispatch(value) {
    return new RemoteReferenceAction(this.lookup(value));
  };

  /**
   * Get an {@link RemoteValue} instance for the given reference value.
   *
   * @param {Reference} reference The remote reference
   * @throws {ReferenceError} If the context not {@link RemoteContext#has} the given reference
   * @return {RemoteValue}
   */


  RemoteContext.prototype.getTarget = function getTarget(reference) {
    return (0, _RemoteValue.reveal)(this.get(reference));
  };

  /**
   * @override
   */


  RemoteContext.prototype.set = function set(reference, value) {
    if (!this[kSession]) {
      throw new TypeError('RemoteSession already closed');
    }

    var remoteValue = new _RemoteValue2.default(this[kSession], reference, value);
    return _ReferenceContext.prototype.set.call(this, reference, remoteValue);
  };

  /**
   * @override
   */


  RemoteContext.prototype.delete = function _delete(reference) {
    this[kFetchRequests].delete(reference);
    if (!_ReferenceContext.prototype.delete.call(this, reference)) return false;

    if (this[kSession] && !this[kSession].isEnded) {
      this[kSession].send(new DeleteAction(reference));
    }

    return true;
  };

  /**
   * @override
   */


  RemoteContext.prototype.release = function release(value) {
    if (!this.parent || !this.parent.exists(value)) {
      (0, _RemoteValue.revoke)(value);
    }

    return _ReferenceContext.prototype.release.call(this, value);
  };

  /**
   * @override
   */


  RemoteContext.prototype.clear = function clear() {
    var _this2 = this;

    this.forEach(function (value, reference) {
      if (_this2[kSession] && !_this2[kSession].isEnded) {
        _this2[kSession].send(new DeleteAction(reference));
      }

      if (!_this2.parent || !_this2.parent.exists(value)) {
        (0, _RemoteValue.revoke)(value);
      }
    });

    this[kFetchRequests].clear();
    _ReferenceContext.prototype.clear.call(this);
  };

  /**
   * End the remote session.
   *
   * @return {RemoteContext}
   */


  RemoteContext.prototype.end = function end() {
    if (this[kSession]) {
      this[kSession].end();
    }

    return this;
  };

  /**
   * Clear the context and destroy the context session only if this context IS NOT created via
   * closure.
   *
   * @return {void}
   */


  RemoteContext.prototype.destroy = function destroy() {
    if (this[kSession]) {
      if (this[kSessionListener]) {
        this[kSession].removeListener('close', this[kSessionListener]);
      }

      if (!(this.parent instanceof RemoteContext)) {
        this[kSession].destroy();
      }
    }

    delete this[kSession];
    delete this[kSessionListener];

    this.clear();
  };

  /**
   * Listen to an event on the remote session.
   *
   * @param {string} event The event to listen to
   * @param {function} listener The listener
   * @return {RemoteContext}
   */


  RemoteContext.prototype.on = function on(event, listener) {
    if (!this[kSession]) {
      throw new TypeError('RemoteSession already closed');
    }

    this[kSession].on(event, listener);
    return this;
  };

  /**
   * Listen once to an event on the remote session.
   *
   * @param {string} event The event to listen to
   * @param {function} listener The listener
   * @return {RemoteContext}
   */


  RemoteContext.prototype.once = function once(event, listener) {
    if (!this[kSession]) {
      throw new TypeError('RemoteSession already closed');
    }

    this[kSession].once(event, listener);
    return this;
  };

  /**
   * Remove listener from the remote session.
   *
   * @param {string} event The event on the listener
   * @param {function} listener The listener
   * @return {RemoteContext}
   */


  RemoteContext.prototype.removeListener = function removeListener(event, listener) {
    if (!this[kSession]) {
      throw new TypeError('RemoteSession already closed');
    }

    this[kSession].removeListener(event, listener);
    return this;
  };

  _createClass(RemoteContext, [{
    key: 'session',
    get: function get() {
      return this[kSession];
    }
  }]);

  return RemoteContext;
}(_referenceContext2.default);

exports.default = RemoteContext;