'use strict';

exports.__esModule = true;

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _remoteProtocol = require('remote-protocol');

var _Context = require('./Context');

var _Context2 = _interopRequireDefault(_Context);

var _LocalContext = require('./LocalContext');

var _LocalContext2 = _interopRequireDefault(_LocalContext);

var _RemoteContext = require('./RemoteContext');

var _RemoteContext2 = _interopRequireDefault(_RemoteContext);

var _actions = require('./actions');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _possibleConstructorReturn(self, call) { if (!self) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return call && (typeof call === "object" || typeof call === "function") ? call : self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function, not " + typeof superClass); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, enumerable: false, writable: true, configurable: true } }); if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass; } /**
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * Copyright 2017 Moshe Simantov
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                *
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * Licensed under the Apache License, Version 2.0 (the "License");
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * you may not use this file except in compliance with the License.
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * You may obtain a copy of the License at
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                *
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                *     http://www.apache.org/licenses/LICENSE-2.0
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                *
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * Unless required by applicable law or agreed to in writing, software
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * distributed under the License is distributed on an "AS IS" BASIS,
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * See the License for the specific language governing permissions and
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * limitations under the License.
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                */

var kLocal = Symbol('local');
var kRemote = Symbol('remote');
var kBinds = Symbol('binds');
var kIsWritable = Symbol('isWritable');
var kShowStack = Symbol('showStack');

/**
 * A remote context {@link Session} to another peer.
 *
 * @extends {Session}
 * @example
 * import { RemoteSession } from 'remote-context';
 *
 * const session = new RemoteSession(objectStream, context);
 */

var RemoteSession = function (_Session) {
  _inherits(RemoteSession, _Session);

  /**
   * Create a session from an object stream.
   *
   * @param {Stream} objectStream an actions object stream
   * @param {Context} context The main context
   * @param {options} [opts] {@link Session} options with extensions
   * @param {boolean} [opts.writable=false] True if this context is writable to the remote peer
   * @param {boolean} [opts.showStack=false] True if errors on this context should include full
   * stack (The stack will be send to the remote peer as a string)
   */
  function RemoteSession(objectStream, context) {
    var opts = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

    _classCallCheck(this, RemoteSession);

    if (!(context instanceof _Context2.default)) {
      throw new TypeError('Invalid context: ' + context);
    }

    /**
     * @type {LocalContext}
     * @private
     */
    var _this = _possibleConstructorReturn(this, _Session.call(this, objectStream, opts));

    _this[kLocal] = new _LocalContext2.default(_this, context);

    /**
     * @type {RemoteContext}
     * @private
     */
    _this[kRemote] = new _RemoteContext2.default(_this, context.parent);

    /**
     * @type {Map}
     * @private
     */
    _this[kBinds] = new Map();

    /**
     * @type {boolean}
     * @private
     */
    _this[kIsWritable] = opts.writable === true;

    /**
     * @type {boolean}
     * @private
     */
    _this[kShowStack] = opts.showStack === true;
    return _this;
  }

  // Session methods

  /**
   * Get the remote context
   *
   * @type {RemoteContext}
   */


  /**
   * Check if the given value is writable.
   * A value is writable if the local-context own it (ie. the value created elusively for this
   * peer) or if the context is writable and the main-context own it. Environment values are
   * always readonly unless there are on the context.
   *
   * @param {*} value The given value
   * @return {boolean}
   */
  RemoteSession.prototype.isWritable = function isWritable(value) {
    if (!this[kLocal]) {
      throw new Error(this + ' already destroyed');
    }

    return this[kLocal].own(value) || this[kIsWritable] && this[kLocal].parent.own(value);
  };

  /**
   * Dispatch a value for the remote peer session.
   * Objects and functions will convert to actions, other primitive values such as string or numbers
   * will remain as-is.
   *
   * @param {*} value
   * @return {Action|null|number|boolean|string}
   */


  RemoteSession.prototype.dispatch = function dispatch(value) {
    switch (typeof value === 'undefined' ? 'undefined' : _typeof(value)) {
      case 'symbol':
      case 'function':
        {
          // break and convert to reference
          break;
        }

      case 'object':
        {
          if (value === null) return null;

          if (value === this || value === this[kLocal] || value === this[kLocal].parent || value === this[kRemote]) {
            throw new ReferenceError("[Safety Check] Can't dispatch internal context instances");
          }

          if (value instanceof _remoteProtocol.Action) {
            throw new TypeError("Bad behaviour, shouldn't dispatch Action instances");
          }

          // break and convert to reference
          break;
        }

      case 'undefined':
      default:
        {
          // `undefined` is not a valid JSON/MsgPack value
          return new _actions.UndefinedValueAction();
        }

      case 'number':
      case 'string':
      case 'boolean':
        {
          return value;
        }
    }

    if (!this[kLocal]) {
      throw new Error(this + ' already destroyed');
    }

    if (this[kRemote].exists(value)) {
      return this[kRemote].dispatch(value);
    }

    return this[kLocal].dispatch(value);
  };

  /**
   * Bind to a local function and apply it remotely if the `this` parameter is a remote value.
   *
   * @param {function} func The local function to bind
   * @return {function} A proxy to the original local function
   */


  RemoteSession.prototype.bind = function bind(func) {
    if (typeof func !== 'function') {
      throw new TypeError('Can\'t bind to non-function value: ' + (typeof func === 'undefined' ? 'undefined' : _typeof(func)));
    }

    if (!this.exists(func)) {
      return func;
    }

    var proxy = this[kBinds].get(func);
    if (!proxy) {
      var self = this;

      proxy = new Proxy(func, {
        apply: function apply(target, thisArg, argumentsList) {
          if (self.exists(thisArg)) {
            return target.apply(thisArg, argumentsList);
          }

          var remote = self.remote;
          var reference = remote.lookup(target);
          return remote.fetch(reference).apply(thisArg, argumentsList);
        }
      });

      this[kBinds].set(func, proxy);
    }

    return proxy;
  };

  /**
   * @override
   */


  RemoteSession.prototype.destroy = function destroy(err) {
    _Session.prototype.destroy.call(this, err);
    if (!this[kLocal]) return;

    this[kLocal].clear();
    this[kRemote].clear();
    this[kBinds].clear();

    delete this[kLocal];
    delete this[kRemote];
    delete this[kBinds];
  };

  // Context methods

  /**
   * Get a local reference value from the local-context.
   *
   * @param {Reference} reference The local reference
   * @return {*}
   */


  RemoteSession.prototype.get = function get(reference) {
    if (!this[kLocal]) {
      throw new Error(this + ' already destroyed');
    }

    return this[kLocal].get(reference);
  };

  /**
   * Set a local reference to the local-context.
   *
   * @param {Reference} reference The local reference
   * @param {*} value The local value
   * @return {RemoteSession}
   */


  RemoteSession.prototype.set = function set(reference, value) {
    if (!this[kLocal]) {
      throw new Error(this + ' already destroyed');
    }

    this[kLocal].set(reference, value);
    return this;
  };

  /**
   * Check if the given value exists on the local-context.
   *
   * @param {*} value The given value
   * @return {boolean} True if exists
   */


  RemoteSession.prototype.exists = function exists(value) {
    if (!this[kLocal]) {
      throw new Error(this + ' already destroyed');
    }

    return this[kLocal].exists(value);
  };

  /**
   * Delete a local reference from the local-context.
   *
   * @param {Reference} reference The local reference
   * @return {boolean} True if the reference has been removed
   */


  RemoteSession.prototype.delete = function _delete(reference) {
    if (!this[kLocal]) {
      throw new Error(this + ' already destroyed');
    }

    return this[kLocal].delete(reference);
  };

  _createClass(RemoteSession, [{
    key: 'remote',
    get: function get() {
      if (!this[kRemote]) {
        throw new Error(this + ' already destroyed');
      }

      return this[kRemote];
    }

    /**
     * True if the remote peer should see stack errors.
     *
     * @type {boolean}
     */

  }, {
    key: 'showStack',
    get: function get() {
      return this[kShowStack];
    }
  }]);

  return RemoteSession;
}(_remoteProtocol.Session);

exports.default = RemoteSession;