'use strict';

exports.__esModule = true;
exports.observe = exports.revoke = exports.reveal = exports.isRemoteValue = undefined;

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _events = require('events');

var _events2 = _interopRequireDefault(_events);

var _RemoteSession = require('./RemoteSession');

var _RemoteSession2 = _interopRequireDefault(_RemoteSession);

var _RemoteContext = require('./RemoteContext');

var _RemoteContext2 = _interopRequireDefault(_RemoteContext);

var _actions = require('./actions');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _possibleConstructorReturn(self, call) { if (!self) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return call && (typeof call === "object" || typeof call === "function") ? call : self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function, not " + typeof superClass); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, enumerable: false, writable: true, configurable: true } }); if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass; } /**
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * Copyright 2017 Moshe Simantov
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                *
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * Licensed under the Apache License, Version 2.0 (the "License");
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * you may not use this file except in compliance with the License.
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * You may obtain a copy of the License at
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                *
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                *     http://www.apache.org/licenses/LICENSE-2.0
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                *
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * Unless required by applicable law or agreed to in writing, software
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * distributed under the License is distributed on an "AS IS" BASIS,
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * See the License for the specific language governing permissions and
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * limitations under the License.
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                */

var kProxyData = Symbol('proxyData');
var kError = Symbol('error');

var proxyMap = new WeakMap();

/**
 * A remote value proxy returned from the {@link RemoteValue} constructor.
 * This proxy make possible to make changes to remote values and get the changes send to
 * the remote peer.
 *
 * @typedef {object|function} RemoteValueProxy
 */

/**
 * Remote value representation locally.
 * @example
 * import { RemoteValue } from 'remote-context';
 *
 * const proxy = new RemoteValue(remoteSession, reference, value);
 */

var RemoteValue = function (_EventEmitter) {
  _inherits(RemoteValue, _EventEmitter);

  /**
   * Check if the give value is a proxy of remote value.
   *
   * @param {RemoteValueProxy|*} proxy The given value
   * @return {boolean}
   */
  RemoteValue.isRemoteValue = function isRemoteValue(proxy) {
    return proxyMap.has(proxy);
  };

  /**
   * Get the remote value instance of the given proxy.
   *
   * @param {RemoteValueProxy} proxy The given proxy
   * @throws {ReferenceError} If the given value is not a {@link RemoteValueProxy}.
   * @return {RemoteValue}
   */


  RemoteValue.reveal = function reveal(proxy) {
    var remoteValue = proxyMap.get(proxy);

    if (!remoteValue) {
      throw new ReferenceError('The given value is not a RemoteValue proxy: ' + proxy);
    }

    return remoteValue;
  };

  /**
   * Resolve the given value.
   * Reject the value if it's a {@link RemoteValueProxy} and an uncaught error accord while
   * using the value (ie. assign property to the value failed). Otherwise, resolve it.
   *
   * @param {RemoteValueProxy|*} value The given value
   * @return {Promise}
   */


  RemoteValue.resolve = function resolve(value) {
    return new Promise(function (resolve, reject) {
      var remoteValue = proxyMap.get(value);
      if (!remoteValue || !remoteValue[kError]) {
        resolve(value);
        return;
      }

      var error = remoteValue[kError];
      remoteValue[kError] = null;

      reject(error);
    });
  };

  /**
   * If the given value is a {@link RemoteValueProxy}, revoke the value proxy.
   * Otherwise, do nothing.
   *
   * @param {*} proxy The given value
   * @return {void}
   */


  RemoteValue.revoke = function revoke(proxy) {
    var remoteValue = proxyMap.get(proxy);
    if (!remoteValue) return;

    remoteValue.revoke();
  };

  /**
   * Observe for any change on the given remote value proxy.
   *
   * @see {@link https://github.com/tc39/proposal-observable}
   * @param {RemoteValueProxy} proxy The given remove value proxy
   * @param {function|null} onNext Called on every change to the given value with the
   * {@link RemoteValueProxy} as the first argument
   * @param {function|null} onError Called on every error occurred on the given value with an
   * error as first argument
   * @param {function|null} onComplete Called when the given value revoked with no arguments
   * @return {{unsubscribe: (function()), closed: boolean}} An subscription object with
   * `unsubscribe` method to stop observer events from calling and a `closed` property.
   */


  RemoteValue.observe = function observe(proxy) {
    var onNext = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : null;
    var onError = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : null;
    var onComplete = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : null;

    var remoteValue = RemoteValue.reveal(proxy);

    var subscription = void 0;
    var events = {
      change: onNext && function (value) {
        return onNext(value);
      },
      error: onError,
      revoke: function revoke() {
        subscription.closed = true;

        if (onComplete !== null) onComplete();
      }
    };

    subscription = {
      unsubscribe: function unsubscribe() {
        Object.keys(events).forEach(function (event) {
          var listener = events[event];
          if (!listener) return;

          remoteValue.removeListener(event, listener);
        });
      },

      closed: !remoteValue.target
    };

    if (!subscription.closed) {
      Object.keys(events).forEach(function (event) {
        var listener = events[event];
        if (listener === null) return;

        remoteValue.on(event, listener);
      });
    }

    return subscription;
  };

  /**
   * Create a {@link RemoteValue} class and a proxy to the given value.
   *
   * @param {RemoteSession} session The current value remote session
   * @param {Reference} reference he value reference for further updates to the remote peer
   * @param {*} value The given value
   * @return {RemoteValueProxy} A proxy to the given value
   */


  function RemoteValue(session, reference, value) {
    var _ret2;

    _classCallCheck(this, RemoteValue);

    if (!(session instanceof _RemoteSession2.default)) {
      throw new TypeError('Expect session to be instance of RemoteSession: ' + session);
    }
    if (!_RemoteContext2.default.isValidReference(reference)) {
      throw new TypeError('The given reference is not valid: ' + reference);
    }

    if (!(value instanceof Object)) {
      var _ret;

      return _ret = value, _possibleConstructorReturn(_this, _ret);
    }

    var _this = _possibleConstructorReturn(this, _EventEmitter.call(this));

    var self = _this;
    _this[kError] = null;

    _this[kProxyData] = Proxy.revocable(value, {
      // FIXME handle property get cache

      get: function get(target, property) {
        var val = target[property];

        if (typeof val !== 'function') {
          return val;
        }

        return session.bind(val);
      },
      setPrototypeOf: function setPrototypeOf(target, prototype) {
        session.request(_actions.LocalSetPrototypeOfAction.fromRemote(session, reference, prototype), function () {}, function (err) {
          return self.reject(err);
        });

        return true;
      },
      preventExtensions: function preventExtensions() {
        session.request(_actions.LocalPreventExtensionsAction.fromRemote(session, reference), function () {}, function (err) {
          return self.reject(err);
        });

        return true;
      },
      defineProperty: function defineProperty(target, property, descriptor) {
        session.request(_actions.LocalDefinePropertyAction.fromRemote(session, reference, property, descriptor), function () {}, function (err) {
          return self.reject(err);
        });

        return true;
      },
      deleteProperty: function deleteProperty(target, property) {
        session.request(_actions.LocalDeletePropertyAction.fromRemote(session, reference, property), function () {}, function (err) {
          return self.reject(err);
        });

        return true;
      }
    });

    _this[kProxyData].target = value;
    var proxy = _this[kProxyData].proxy;


    proxyMap.set(proxy, _this);
    return _ret2 = proxy, _possibleConstructorReturn(_this, _ret2);
  }

  RemoteValue.prototype.reject = function reject(error) {
    if (!this[kError]) {
      this[kError] = error;
    }

    try {
      this.emit('error', error);
    } catch (err) {
      // Ignore throwing error if no listeners found
    }
  };

  RemoteValue.prototype.revoke = function revoke() {
    delete this[kProxyData].target;
    this[kProxyData].revoke();

    this.emit('revoke');
  };

  // Update methods

  RemoteValue.prototype.setPrototypeOf = function setPrototypeOf(prototype) {
    var _kProxyData = this[kProxyData],
        target = _kProxyData.target,
        proxy = _kProxyData.proxy;

    if (!target) return false;

    Object.setPrototypeOf(target, prototype);
    this.emit('change', proxy, 'setPrototypeOf');

    return true;
  };

  RemoteValue.prototype.preventExtensions = function preventExtensions() {
    var _kProxyData2 = this[kProxyData],
        target = _kProxyData2.target,
        proxy = _kProxyData2.proxy;

    if (!target) return false;

    Object.preventExtensions(target);
    this.emit('change', proxy, 'preventExtensions');

    return true;
  };

  RemoteValue.prototype.defineProperty = function defineProperty(property, descriptor) {
    var _kProxyData3 = this[kProxyData],
        target = _kProxyData3.target,
        proxy = _kProxyData3.proxy;

    if (!target) return false;

    Object.defineProperty(target, property, descriptor);
    this.emit('change', proxy, 'defineProperty', property);

    return true;
  };

  RemoteValue.prototype.deleteProperty = function deleteProperty(property) {
    var _kProxyData4 = this[kProxyData],
        target = _kProxyData4.target,
        proxy = _kProxyData4.proxy;

    if (!target) return false;

    Object.deleteProperty(target, property);
    this.emit('change', proxy, 'deleteProperty', property);

    return true;
  };

  _createClass(RemoteValue, [{
    key: 'proxy',
    get: function get() {
      return this[kProxyData].proxy;
    }
  }, {
    key: 'target',
    get: function get() {
      return this[kProxyData].target;
    }
  }]);

  return RemoteValue;
}(_events2.default);

/**
 * @type {RemoteValue.isRemoteValue}
 */


exports.default = RemoteValue;
var isRemoteValue = exports.isRemoteValue = RemoteValue.isRemoteValue;

/**
 * @type {RemoteValue.reveal}
 */
var reveal = exports.reveal = RemoteValue.reveal;

/**
 * @type {RemoteValue.revoke}
 */
var revoke = exports.revoke = RemoteValue.revoke;

/**
 * @type {RemoteValue.observe}
 */
var observe = exports.observe = RemoteValue.observe;