'use strict';

exports.__esModule = true;

var _remoteProtocol = require('remote-protocol');

var _actions = require('./actions');

/**
 * A map with all the context actions with their's codes.
 *
 * @type {object}
 */
/**
 * Copyright 2017 Moshe Simantov
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

exports.default = {
  // WARNING: ORDER IS MATTER!
  // Children classes needs to be before parents

  // Request-response actions
  0x01: _remoteProtocol.RequestAction,
  0x02: _remoteProtocol.ResponseAction,

  // Context actions
  0x10: _actions.GetAction,
  0x11: _actions.RemoteSetFunctionAction,
  0x12: _actions.RemoteSetObjectAction,
  0x13: _actions.RemoteSetSymbolAction,
  0x14: _actions.DeleteAction,
  0x15: _actions.LocalReferenceAction,
  0x16: _actions.RemoteReferenceAction,

  // Local Actions
  0x20: _actions.LocalDefinePropertyAction,
  0x21: _actions.LocalDeletePropertyAction,
  0x22: _actions.LocalSetPrototypeOfAction,
  0x23: _actions.LocalPreventExtensionsAction,

  // Remote Actions
  0x30: _actions.RemoteDefinePropertyAction,
  0x31: _actions.RemoteDeletePropertyAction,
  0x32: _actions.RemoteSetPrototypeOfAction,
  0x33: _actions.RemotePreventExtensionsAction,

  // Reflect Actions
  0x40: _actions.ReflectApplyAction,
  0x41: _actions.ReflectConstructAction,
  0x42: _actions.ReflectGetAction,
  0x43: _actions.ReflectPromiseAction,

  // Custom types
  0x51: _actions.PropertyDescriptorAction,
  0x52: _actions.PropertyDescriptorsAction,
  0x53: _actions.UndefinedValueAction
};